@@ -9,15 +9,15 @@ task queue situations where tasks are small, idempotent and run very fast.
 
 SQS Features supported by this transport:
   Long Polling:
-    http://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/
+    https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/
       sqs-long-polling.html
 
     Long polling is enabled by setting the `wait_time_seconds` transport
     option to a number > 1.  Amazon supports up to 20 seconds.  This is
-    disabled for now, but will be enabled by default in the near future.
+    enabled with 10 seconds by default.
 
   Batch API Actions:
-   http://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/
+   https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/
      sqs-batch-api.html
 
     The default behavior of the SQS Channel.drain_events() method is to
@@ -37,17 +37,17 @@ SQS Features supported by this transport:
 
 from __future__ import absolute_import, unicode_literals
 
+import base64
 import socket
 import string
+import uuid
 
 from vine import transform, ensure_promise, promise
 
-from kombu.async import get_event_loop
-from kombu.async.aws import sqs as _asynsqs
-from kombu.async.aws.ext import boto, exception
-from kombu.async.aws.sqs.connection import AsyncSQSConnection, SQSConnection
-from kombu.async.aws.sqs.ext import regions
-from kombu.async.aws.sqs.message import Message
+from kombu.asynchronous import get_event_loop
+from kombu.asynchronous.aws.ext import boto3, exceptions
+from kombu.asynchronous.aws.sqs.connection import AsyncSQSConnection
+from kombu.asynchronous.aws.sqs.message import AsyncMessage
 from kombu.five import Empty, range, string_t, text_t
 from kombu.log import get_logger
 from kombu.utils import scheduling
@@ -83,7 +83,7 @@ class Channel(virtual.Channel):
 
     default_region = 'us-east-1'
     default_visibility_timeout = 1800  # 30 minutes.
-    default_wait_time_seconds = 10  # disabled see #198
+    default_wait_time_seconds = 10  # up to 20 seconds max
     domain_format = 'kombu%(vhost)s'
     _asynsqs = None
     _sqs = None
@@ -91,8 +91,8 @@ class Channel(virtual.Channel):
     _noack_queues = set()
 
     def __init__(self, *args, **kwargs):
-        if boto is None:
-            raise ImportError('boto is not installed')
+        if boto3 is None:
+            raise ImportError('boto3 is not installed')
         super(Channel, self).__init__(*args, **kwargs)
 
         # SQS blows up if you try to create a new queue when one already
@@ -104,18 +104,10 @@ class Channel(virtual.Channel):
         self.hub = kwargs.get('hub') or get_event_loop()
 
     def _update_queue_cache(self, queue_name_prefix):
-        try:
-            queues = self.sqs.get_all_queues(prefix=queue_name_prefix)
-        except exception.SQSError as exc:
-            if exc.status == 403:
-                raise RuntimeError(
-                    'SQS authorization error, access_key={0}'.format(
-                        self.sqs.access_key))
-            raise
-        else:
-            self._queue_cache.update({
-                queue.name: queue for queue in queues
-            })
+        resp = self.sqs.list_queues(QueueNamePrefix=queue_name_prefix)
+        for url in resp.get('QueueUrls', []):
+            queue_name = url.split('/')[-1]
+            self._queue_cache[queue_name] = url
 
     def basic_consume(self, queue, no_ack, *args, **kwargs):
         if no_ack:
@@ -132,7 +124,7 @@ class Channel(virtual.Channel):
             self._noack_queues.discard(queue)
         return super(Channel, self).basic_cancel(consumer_tag)
 
-    def drain_events(self, timeout=None):
+    def drain_events(self, timeout=None, callback=None, **kwargs):
         """Return a single payload message from one of our queues.
 
         Raises:
@@ -143,7 +135,7 @@ class Channel(virtual.Channel):
             raise Empty()
 
         # At this point, go and get more messages from SQS
-        self._poll(self.cycle, self.connection._deliver, timeout=timeout)
+        self._poll(self.cycle, callback, timeout=timeout)
 
     def _reset_cycle(self):
         """Reset the consume cycle.
@@ -160,7 +152,15 @@ class Channel(virtual.Channel):
 
     def entity_name(self, name, table=CHARS_REPLACE_TABLE):
         """Format AMQP queue name into a legal SQS queue name."""
-        return text_t(safe_str(name)).translate(table)
+        if name.endswith('.fifo'):
+            partial = name.rstrip('.fifo')
+            partial = text_t(safe_str(partial)).translate(table)
+            return partial + '.fifo'
+        else:
+            return text_t(safe_str(name)).translate(table)
+
+    def canonical_queue_name(self, queue_name):
+        return self.entity_name(self.queue_name_prefix + queue_name)
 
     def _new_queue(self, queue, **kwargs):
         """Ensure a queue with given name exists in SQS."""
@@ -168,7 +168,7 @@ class Channel(virtual.Channel):
             return queue
         # Translate to SQS name for consistency with initial
         # _queue_cache population.
-        queue = self.entity_name(self.queue_name_prefix + queue)
+        queue = self.canonical_queue_name(queue)
 
         # The SQS ListQueues method only returns 1000 queues.  When you have
         # so many queues, it's possible that the queue you are looking for is
@@ -179,10 +179,14 @@ class Channel(virtual.Channel):
         try:
             return self._queue_cache[queue]
         except KeyError:
-            q = self._queue_cache[queue] = self.sqs.create_queue(
-                queue, self.visibility_timeout,
-            )
-            return q
+            attributes = {'VisibilityTimeout': str(self.visibility_timeout)}
+            if queue.endswith('.fifo'):
+                attributes['FifoQueue'] = 'true'
+
+            resp = self._queue_cache[queue] = self.sqs.create_queue(
+                QueueName=queue, Attributes=attributes)
+            self._queue_cache[queue] = resp['QueueUrl']
+            return resp['QueueUrl']
 
     def _delete(self, queue, *args, **kwargs):
         """Delete queue by name."""
@@ -191,15 +195,28 @@ class Channel(virtual.Channel):
 
     def _put(self, queue, message, **kwargs):
         """Put message onto queue."""
-        q = self._new_queue(queue)
-        m = Message()
-        m.set_body(dumps(message))
-        q.write(m)
+        q_url = self._new_queue(queue)
+        kwargs = {'QueueUrl': q_url,
+                  'MessageBody': AsyncMessage().encode(dumps(message))}
+        if queue.endswith('.fifo'):
+            if 'MessageGroupId' in message['properties']:
+                kwargs['MessageGroupId'] = \
+                    message['properties']['MessageGroupId']
+            else:
+                kwargs['MessageGroupId'] = 'default'
+            if 'MessageDeduplicationId' in message['properties']:
+                kwargs['MessageDeduplicationId'] = \
+                    message['properties']['MessageDeduplicationId']
+            else:
+                kwargs['MessageDeduplicationId'] = str(uuid.uuid4())
+        self.sqs.send_message(**kwargs)
 
     def _message_to_python(self, message, queue_name, queue):
-        payload = loads(bytes_to_str(message.get_body()))
+        body = base64.b64decode(message['Body'].encode())
+        payload = loads(bytes_to_str(body))
         if queue_name in self._noack_queues:
-            queue.delete_message(message)
+            queue = self._new_queue(queue_name)
+            self.asynsqs.delete_message(queue, message['ReceiptHandle'])
         else:
             try:
                 properties = payload['properties']
@@ -209,14 +226,14 @@ class Channel(virtual.Channel):
                 delivery_info = {}
                 properties = {'delivery_info': delivery_info}
                 payload.update({
-                    'body': bytes_to_str(message.get_body()),
+                    'body': bytes_to_str(body),
                     'properties': properties,
                 })
-        # set delivery tag to SQS receipt handle
-        delivery_info.update({
-            'sqs_message': message, 'sqs_queue': queue,
-        })
-        properties['delivery_tag'] = message.receipt_handle
+            # set delivery tag to SQS receipt handle
+            delivery_info.update({
+                'sqs_message': message, 'sqs_queue': queue,
+            })
+            properties['delivery_tag'] = message['ReceiptHandle']
         return payload
 
     def _messages_to_python(self, messages, queue):
@@ -261,23 +278,32 @@ class Channel(virtual.Channel):
         # drain_events calls `can_consume` first, consuming
         # a token, so we know that we are allowed to consume at least
         # one message.
-        maxcount = self._get_message_estimate()
-        if maxcount:
-            q = self._new_queue(queue)
-            messages = q.get_messages(num_messages=maxcount)
 
-            if messages:
-                for msg in self._messages_to_python(messages, queue):
+        # Note: ignoring max_messages for SQS with boto3
+        max_count = self._get_message_estimate()
+        if max_count:
+            q_url = self._new_queue(queue)
+            resp = self.sqs.receive_message(
+                QueueUrl=q_url, MaxNumberOfMessages=max_count,
+                WaitTimeSeconds=self.wait_time_seconds)
+            if resp.get('Messages'):
+                for m in resp['Messages']:
+                    m['Body'] = AsyncMessage(body=m['Body']).decode()
+                for msg in self._messages_to_python(resp['Messages'], queue):
                     self.connection._deliver(msg, queue)
                 return
         raise Empty()
 
     def _get(self, queue):
         """Try to retrieve a single message off ``queue``."""
-        q = self._new_queue(queue)
-        messages = q.get_messages(num_messages=1)
-        if messages:
-            return self._messages_to_python(messages, queue)[0]
+        q_url = self._new_queue(queue)
+        resp = self.sqs.receive_message(
+            QueueUrl=q_url, MaxNumberOfMessages=1,
+            WaitTimeSeconds=self.wait_time_seconds)
+        if resp.get('Messages'):
+            body = AsyncMessage(body=resp['Messages'][0]['Body']).decode()
+            resp['Messages'][0]['Body'] = body
+            return self._messages_to_python(resp['Messages'], queue)[0]
         raise Empty()
 
     def _loop1(self, queue, _=None):
@@ -311,17 +337,18 @@ class Channel(virtual.Channel):
 
     def _get_async(self, queue, count=1, callback=None):
         q = self._new_queue(queue)
+        qname = self.canonical_queue_name(queue)
         return self._get_from_sqs(
-            q, count=count, connection=self.asynsqs,
+            qname, count=count, connection=self.asynsqs,
             callback=transform(self._on_messages_ready, callback, q, queue),
         )
 
     def _on_messages_ready(self, queue, qname, messages):
-        if messages:
+        if 'Messages' in messages and messages['Messages']:
             callbacks = self.connection._callbacks
-            for raw_message in messages:
-                message = self._message_to_python(raw_message, qname, queue)
-                callbacks[qname](message)
+            for msg in messages['Messages']:
+                msg_parsed = self._message_to_python(msg, qname, queue)
+                callbacks[qname](msg_parsed)
 
     def _get_from_sqs(self, queue,
                       count=1, connection=None, callback=None):
@@ -344,72 +371,69 @@ class Channel(virtual.Channel):
         return super(Channel, self)._restore(message)
 
     def basic_ack(self, delivery_tag, multiple=False):
-        delivery_info = self.qos.get(delivery_tag).delivery_info
         try:
-            queue = delivery_info['sqs_queue']
+            message = self.qos.get(delivery_tag).delivery_info
+            sqs_message = message['sqs_message']
         except KeyError:
             pass
         else:
-            queue.delete_message(delivery_info['sqs_message'])
+            self.asynsqs.delete_message(message['sqs_queue'],
+                                        sqs_message['ReceiptHandle'])
         super(Channel, self).basic_ack(delivery_tag)
 
     def _size(self, queue):
         """Return the number of messages in a queue."""
-        return self._new_queue(queue).count()
+        url = self._new_queue(queue)
+        resp = self.sqs.get_queue_attributes(
+            QueueUrl=url,
+            AttributeNames=['ApproximateNumberOfMessages'])
+        return int(resp['Attributes']['ApproximateNumberOfMessages'])
 
     def _purge(self, queue):
         """Delete all current messages in a queue."""
         q = self._new_queue(queue)
         # SQS is slow at registering messages, so run for a few
-        # iterations to ensure messages are deleted.
+        # iterations to ensure messages are detected and deleted.
         size = 0
         for i in range(10):
-            size += q.count()
+            size += int(self._size(queue))
             if not size:
                 break
-        q.clear()
+        self.sqs.purge_queue(QueueUrl=q)
         return size
 
     def close(self):
         super(Channel, self).close()
-        for conn in (self._sqs, self._asynsqs):
-            if conn:
-                try:
-                    conn.close()
-                except AttributeError as exc:  # FIXME ???
-                    if "can't set attribute" not in str(exc):
-                        raise
-
-    def _get_regioninfo(self, regions):
-        if self.regioninfo:
-            return self.regioninfo
-        if self.region:
-            for _r in regions:
-                if _r.name == self.region:
-                    return _r
-
-    def _aws_connect_to(self, fun, regions):
-        conninfo = self.conninfo
-        region = self._get_regioninfo(regions)
-        is_secure = self.is_secure if self.is_secure is not None else True
-        port = self.port if self.port is not None else conninfo.port
-        return fun(region=region,
-                   aws_access_key_id=conninfo.userid,
-                   aws_secret_access_key=conninfo.password,
-                   is_secure=is_secure,
-                   port=port)
+        # if self._asynsqs:
+        #     try:
+        #         self.asynsqs.close()
+        #     except AttributeError as exc:  # FIXME ???
+        #         if "can't set attribute" not in str(exc):
+        #             raise
 
     @property
     def sqs(self):
         if self._sqs is None:
-            self._sqs = self._aws_connect_to(SQSConnection, regions())
+            session = boto3.session.Session(
+                region_name=self.region,
+                aws_access_key_id=self.conninfo.userid,
+                aws_secret_access_key=self.conninfo.password,
+            )
+            is_secure = self.is_secure if self.is_secure is not None else True
+            client_kwargs = {
+                'use_ssl': is_secure
+            }
+            if self.endpoint_url is not None:
+                client_kwargs['endpoint_url'] = self.endpoint_url
+            self._sqs = session.client('sqs', **client_kwargs)
         return self._sqs
 
     @property
     def asynsqs(self):
         if self._asynsqs is None:
-            self._asynsqs = self._aws_connect_to(
-                AsyncSQSConnection, _asynsqs.regions(),
+            self._asynsqs = AsyncSQSConnection(
+                sqs_connection=self.sqs,
+                region=self.region
             )
         return self._asynsqs
 
@@ -450,6 +474,20 @@ class Channel(virtual.Channel):
     def port(self):
         return self.transport_options.get('port')
 
+    @cached_property
+    def endpoint_url(self):
+        if self.conninfo.hostname is not None:
+            scheme = 'https' if self.is_secure else 'http'
+            if self.conninfo.port is not None:
+                port = ':{}'.format(self.conninfo.port)
+            else:
+                port = ''
+            return '{}://{}{}'.format(
+                scheme,
+                self.conninfo.hostname,
+                port
+            )
+
     @cached_property
     def wait_time_seconds(self):
         return self.transport_options.get('wait_time_seconds',
@@ -466,15 +504,19 @@ class Transport(virtual.Transport):
     default_port = None
     connection_errors = (
         virtual.Transport.connection_errors +
-        (exception.SQSError, socket.error)
+        (exceptions.BotoCoreError, socket.error)
     )
     channel_errors = (
-        virtual.Transport.channel_errors + (exception.SQSDecodeError,)
+        virtual.Transport.channel_errors + (exceptions.BotoCoreError,)
     )
     driver_type = 'sqs'
     driver_name = 'sqs'
 
     implements = virtual.Transport.implements.extend(
-        async=True,
+        asynchronous=True,
         exchange_type=frozenset(['direct']),
     )
+
+    @property
+    def default_connection_params(self):
+        return {'port': self.default_port}
